from hashlib import sha256
from jose import jwt
from pydantic import Field, BaseModel
from typing import Optional


class DBCredentials(BaseModel):
    database_url: str
    token: Optional[str] = None
    
    def hash(self):
        return sha256(self.database_url.encode()).hexdigest()
    
    def encode(self):
        return jwt.encode(self.database_url, self.hash())
    
    def decode(self):
        return jwt.decode(self.database_url, self.hash())
    
    
    def __init__(self, **data):
        super().__init__(**data)
        self.token = jwt.encode(
            {"database_url": self.database_url},
            self.hash()
        )