from odmantic import Model, Field as oField
from pydantic import BaseModel, EmailStr, Field
from typing import Optional, List
from datetime import datetime
from jose import jwt
from hashlib import sha256
from json import loads
from odmantic import Model, ObjectId, Field as oField
from pydantic import BaseModel, EmailStr, Field
from typing import Optional, List
from datetime import datetime
from jose import jwt
from hashlib import sha256
from fastapi import FastAPI, Request, Response, HTTPException, Depends, status, Body
from fastapi.middleware.cors import CORSMiddleware
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from odmantic import AIOEngine
from motor.motor_asyncio import AsyncIOMotorClient as Async
from fastapi_lambda.config import Settings
from fastapi.responses import JSONResponse


class User(Model):
    email: EmailStr = oField(...)
    password: str = oField(...)
    token: Optional[str] = oField(default=None)

    def token_(self):
        return jwt.encode(
            {"email": self.email, "password": self.password},
            sha256(self.password.encode()).hexdigest(),
        )


class UserSchema(BaseModel):
    email: EmailStr = Field(...)
    password: str = Field(...)
    token: Optional[str] = Field(default=None)

    def __init__(self, **data):
        super().__init__(**data)
        self.password = sha256(self.password.encode()).hexdigest()
        self.token = jwt.encode(
            {"email": self.email, "password": data["password"]}, self.password
        )


class RequestLog(BaseModel):
    """Request log Schema model"""

    method: str = Field(...)
    url: str = Field(...)
    status_code: int = Field(...)
    headers: dict = Field(...)
    time_data: str = Field(
        default_factory=lambda: datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    )


class RequestLogModel(Model):
    """Request log Database model"""

    method: str = oField(...)
    url: str = oField(...)
    status_code: int = oField(...)
    headers: dict = oField(...)
    time_data: str = oField(...)


class ServerlessFastAPI(FastAPI):
    """FastAPI Serverless class"""

    def __init__(self, mongo_url: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.title = "FastAPI Serverless"
        self.description = "A Wrselfer for FastAPI to run on AWS Lambda"
        self.version = "0.0.1"
        self.oauth2_scheme = OAuth2PasswordBearer(tokenUrl="oauth2/token")
        self.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        if mongo_url is None:
            mongo_url = Settings().DATABASE_URL
        self.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        self.title = "FastAPI Serverless"
        self.description = "FastAPI Serverless"
        self.version = "0.0.1"
        self.db = AIOEngine(
            client=Async(mongo_url), database=mongo_url.split("/")[-1].split("?")[0]
        )

        @self.middleware("http")
        async def log_requests(request: Request, call_next):
            """Logins requests middleware"""
            try:
                if request.url.path in [
                    "/docs",
                    "/docs/dashboard",
                    "/redoc",
                    "/openapi.json",
                    "/favicon.ico",
                ]:
                    return await call_next(request)
                elif request.url.hostname == "localhost":
                    return await call_next(request)
                else:
                    response = await call_next(request)
                    body = b""
                    headers = dict(request.headers)
                    async for chunk in response.body_iterator:
                        body += chunk
                    log = RequestLog(
                        method=request.method,
                        url=str(request.url),
                        status_code=response.status_code,
                        headers=headers,
                    )
                    await self.db.save(RequestLogModel(**log.dict()))
                    return Response(
                        content=body,
                        status_code=response.status_code,
                        headers=response.headers,
                    )
            except HTTPException as exc:
                print(exc)
                return await call_next(request)

        @self.get("/docs/dashboard", include_in_schema=False)
        async def dashboard_api():
            """Logs of all requests"""
            instances = await self.db.find(RequestLogModel)
            response = [loads(instance.json()) for instance in instances]
            return JSONResponse(response)

        @self.post("/oauth2/token", response_class=JSONResponse)
        async def login_for_access_token(
            form_data: OAuth2PasswordRequestForm = Depends(),
        ):
            user = await self.db.find_one(User, User.email == form_data.username)
            if not user:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail="Incorrect email or password",
                    headers={"WWW-Authenticate": "Bearer"},
                )
            if user.password != sha256(form_data.password.encode()).hexdigest():
                return JSONResponse(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    content={"detail": "Incorrect email or password"},
                )
            return {"access_token": user.token, "token_type": "bearer"}

        @self.get("/oauth2/userinfo")
        async def read_users_me(token: str = Depends(self.oauth2_scheme)):
            try:
                user = await self.db.find_one(User, User.token == token)
                return jwt.decode(token, user.password, algorithms=["HS256"])
            except Exception as exc:
                print(exc)
                return JSONResponse(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    content={"detail": "Incorrect email or password"},
                )

        @self.post("/oauth2/authorize")
        async def register(body=Body(...)):
            """ **The request body must have the following fields:**
            - 📧 email: string
            - 🔒 password: string"""
            user_dict = loads(body)
            if await self.db.find_one(User, User.email == user_dict["email"]):
                return JSONResponse(
                    status_code=status.HTTP_409_CONFLICT,
                    content={"detail": "User already exists"},
                )
            user = UserSchema(**user_dict)
            await self.db.save(User(**user.dict()))
            return {"message": "User created successfully"}